'''
Contains all you need to queue up a sequence of actions to be carried out in the future.
'''
from threading import Thread, RLock
from Queue import Queue
import time
import logging


class Sequencer(Thread):
    '''
    A threaded sequencer that queues Events and executes them periodically. You should call start() once you've
    created your Sequencer: this isn't done for you. The queue is checked, and events executed, at an interval
    of around one second. This means that there will be a delay of around one second between two consecutive
    scheduled Events - there is no need to add a SleepEvent(1) between them.

    Once started, the Sequencer runs in a daemonized thread and will keep running until your program otherwise
    terminates.

    Events can call any Python function, but in this context are probably calls to methods on a Controller, or the
    special 'wait' SleepEvent, which does as it says on the tin.

    Note that Pyro doesn't pickle functions, so if you're sequencing something client-side you'll need to use the
    ControllerEvent or DeviceEvent classes, and you'll only have access to the functions on the Controller and devices
    you'd call normally with Pyro.

    There should probably only be one Sequencer in a system, and it should probably reside on the controller side
    and not the user interface side. Breaching either of these conditions should be preceded by some Deep Thought
    about possible side effects.
    '''

    def __init__(self, controller):
        '''Initialise, but do not start(), the Sequencer.'''
        Thread.__init__(self)
        self.controller = controller
        self.queue = Queue()
        self.setDaemon(True)
        self.lock = RLock()

    def run(self):
        logging.info("Starting sequencer loop")
        while 1:
            time.sleep(1)
            event = self.queue.get()
            event.execute(self.controller)

    def sequence(self, *events):
        '''Add one or more Events to the sequencer queue. Events added in one call are guaranteed to be executed in order
        and without interleaving with other event sequences.'''
        with self.lock:
            for event in events:
                self.queue.put(event)


class Event(object):
    '''An Event is essentially a preserved method call, to be executed at an unspecified point in the future. This will NOT work over
    Pyro RPC: use a ControllerEvent with a named method argument instead.'''

    def __init__(self, method, *args):
        '''Create an Event that, when executed, will call the given method and provide the given arguments.
        Keyword arguments are not (currently) supported.'''
        self.method = method
        self.args = args

    def execute(self, controller):
        '''Execute the Event, by calling the relevant method with the given arguments.'''
        self.method(*self.args)


class ControllerEvent(Event):
    '''A ControllerEvent invokes a method on the Controller. The method is supplied by name.'''

    def __init__(self, method, *args):
        Event.__init__(self, method, *args)

    def execute(self, controller):
        getattr(controller, self.method)(*self.args)


class DeviceEvent(Event):
    '''A DeviceEvent invokes a method on a device on the Controller. The device ID and method are supplied by name.'''

    def __init__(self, deviceID, method, *args):
        Event.__init__(self, method, *args)
        self.deviceID = deviceID

    def execute(self, controller):
        getattr(controller.getDevice(self.deviceID), self.method)(*self.args)


class SleepEvent(Event):
    '''Sleep for a certain number of seconds.'''

    def __init__(self, seconds):
        Event.__init__(self, None)
        self.seconds = seconds

    def execute(self, controller):
        logging.debug("Sleeping sequencer for " + str(self.seconds) + " seconds")
        time.sleep(self.seconds)


class CompositeEvent(object):
    '''An Event that immediately triggers any of its child events.'''

    def __init__(self, *events):
        self.events = events

    def addEvent(self, event):
        self.events.append(event)

    def execute(self, controller):
        for event in self.events:
            event.execute(controller)


class LogEvent(object):
    '''An Event that writes to the controller's log.'''

    def __init__(self, level, msg):
        self.level = level
        self.msg = msg

    def execute(self, controller):
        logging.log(self.level, self.msg)


class BroadcastEvent(object):
    def __init__(self, msgType, source, data):
        self.msgType = msgType
        self.source = source
        self.data = data

    def execute(self, controller):
        controller.broadcast(self.msgType, self.source, self.data)



from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
from threading import Thread
import logging


def handleRequest(path, controller, respond):
    if path.count('/') is 3:
        _, deviceID, method, args = path.split('/')
        if controller.hasDevice(deviceID):
            device = controller.getDevice(deviceID)
            if device.httpAccessible:
                function = getattr(device, method)
                if function:
                    result = function(*args.split(','))
                    respond(200, "OK " + str(result))
                else:
                    respond(400, "No such function: " + method)
            else:
                respond(403, "Not permitted to access device " + deviceID + " over HTTP.")
        else:
            respond(404, "No such device: " + deviceID)
    else:
        respond(400, "Incorrectly formatted URL: " + path)


class ControllerHttp(Thread):
    '''
    An HTTP interface to a Controller's methods.
    '''

    def __init__(self, controller, port=8080):
        super(ControllerHttp, self).__init__()
        self.controller = controller
        self.port = port
        self.setDaemon(True)
        self.ControllerHttpRequestHandler.controller = controller

    class ControllerHttpRequestHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            handleRequest(self.path, self.controller, self.respond)

        def respond(self, responseCode, text):
            self.send_response(responseCode)
            self.send_header('Content-type', 'text/plain')
            self.end_headers()
            self.wfile.write(text)

    def run(self):
        self.server = HTTPServer(("", self.port), self.ControllerHttpRequestHandler)
        logging.info("Started HTTP server on port " + str(self.port))
        self.server.serve_forever()



import logging


class Device(object):
    '''
    This is a thing. It does stuff.
    '''

    def __init__(self, deviceID, httpAccessible=False, **kwargs):
        self.deviceID = deviceID
        self.httpAccessible = httpAccessible

    def initialise(self):
        '''
        Extension point for device-specific initialisation code (e.g. resetting to known-default config)
        '''
        pass

    def deinitialise(self):
        '''
        Extension point for device-specific deinitialisation code (e.g. closing open file handles)
        '''
        pass

    @staticmethod
    def create(d, controller):
        deviceID = d["deviceID"]
        logging.info("Creating device " + deviceID)
        if "options" in d:
            return get_class(d["class"])(deviceID, controller=controller, **d["options"])
        else:
            return get_class(d["class"])(deviceID, controller=controller)


def get_class(kls):
    parts = kls.split('.')
    module = ".".join(parts[:-1])
    m = __import__(module)
    for comp in parts[1:]:
        m = getattr(m, comp)
    return m


class InvalidArgumentException(Exception):
    pass



from avx.devices import Device
from socket import socket


class Tivo(Device):
    '''
    A networked TiVo device. Developed against Virgin Media UK's TiVo boxes.
    '''

    socket = None

    def __init__(self, deviceID, ipAddress, port=31339, **kwargs):
        super(Tivo, self).__init__(deviceID, **kwargs)
        self.ipAddress = ipAddress
        self.port = port

    def initialise(self):
        self.socket = socket()
        self.socket.settimeout(5)
        self.socket.connect((self.ipAddress, self.port))

    def send(self, message):
        if not self.socket:
            self.initialise()
        self.socket.sendall(message)

    def sendIRCode(self, ircode):
        self.send('IRCODE %s\r' % ircode)

    # Playback functions

    def pause(self):
        self.sendIRCode("PAUSE")

    def play(self):
        self.sendIRCode("PLAY")

    def rewind(self):
        self.sendIRCode("REVERSE")

    def fastForward(self):
        self.sendIRCode("FORWARD")

    def replay(self):
        self.sendIRCode("REPLAY")

    def skip(self):
        self.sendIRCode("ADVANCE")

    def slow(self):
        self.sendIRCode("SLOW")



from avx.Client import MessageTypes
from avx.devices.serial import SerialDevice


class Inline3808(SerialDevice):
    '''
    An Inline 3808 8x4 switcher.
    '''

    def __init__(self, deviceID, serialDevice, **kwargs):
        super(Inline3808, self).__init__(deviceID, serialDevice, baud=1200, **kwargs)

    def initialise(self):
        self.sendCommand("[CNF290000]")  # All boards to group 1. Copying what the AMX does currently though this should be the default.

    def sendInputToOutput(self, inChannel, outChannel):
        return self.sendCommand("[PT1O0" + str(outChannel) + "I0" + str(inChannel) + "]")

    def requestStatus(self):
        self.sendCommand("[VID]")

    def hasFullMessage(self, recv_buffer):
        return recv_buffer[-1] == "]"

    def handleMessage(self, msgBytes):
        msgString = ''.join(map(chr, msgBytes))
        if len(msgString) == 13 and msgString[1:4] == "VID":
            self.broadcast(
                MessageTypes.OUTPUT_MAPPING,
                {
                    1: int(msgString[5]),
                    2: int(msgString[7]),
                    3: int(msgString[9]),
                    4: int(msgString[11]),
                }
            )



from avx.devices.serial import SerialDevice


max_command_length = 200


def wrapUnisonCommand(func):
    '''
    Decorator for a function which should return the USAP command string to be sent.
    The resulting function requires `self.sendCommand` to resolve to an appropriate function.
    '''
    def unisonCommandInner(self, *args):
        cmd = UnisonCommand(func(self, *args))
        return self.sendCommand(cmd.getByteString())
    return unisonCommandInner


class UnisonDevice(SerialDevice):
    '''
    An ETC Unison lighting controller speaking Unison Serial Access Protocol (USAP).
    '''

    def __init__(self, deviceID, serialDevice, **kwargs):
        super(UnisonDevice, self).__init__(deviceID, serialDevice, **kwargs)

    # Commands for presets

    @wrapUnisonCommand
    def activate(self, unisonObject):
        return "{}.ACTI".format(unisonObject)

    @wrapUnisonCommand
    def activateWithFade(self, preset, fadeTime):
        return "{}.nDFT={}".format(preset, fadeTime)

    @wrapUnisonCommand
    def deactivate(self, unisonObject):
        return "{}.DACT".format(unisonObject)

    # Commands for walls

    @wrapUnisonCommand
    def open(self, wall):
        return "{}.OPEN".format(wall)

    @wrapUnisonCommand
    def close(self, wall):
        return "{}.CLOS".format(wall)

    @wrapUnisonCommand
    def toggleOpen(self, wall):
        return "{}.TOGL".format(wall)

    # Commands for zones

    @wrapUnisonCommand
    def setZoneIntensity(self, zone, intensity):
        # 0 <= intensity <= 100, command needs a percentage of 65535
        return "{}.nINT={:0.0f}".format(zone, intensity * 655.35)

    # Commands for macros

    @wrapUnisonCommand
    def execute(self, macroName):
        return "{}.EXEC".format(macroName)

    @wrapUnisonCommand
    def stop(self, macroName):
        return "{}.STOP".format(macroName)

    # Commands for section masters

    @wrapUnisonCommand
    def master(self, section, level):
        return "{}.Master.nVAL={:0.0f}".format(section, level * 655.35)


class UnisonCommand(object):
    '''
    Wrapper around a string command to add the binary bits to be sent.
    '''

    def __init__(self, command):
        self.checkLength(command)
        self.command = command

    def getByteString(self):
        self.checkLength(self.command)
        return '\xEE' + chr(len(self.command) + 3) + '\x00\x00\x40' + self.command + '\x00\x00'

    def checkLength(self, cmdString):
        cmdLength = len(cmdString)
        if cmdLength > max_command_length:
            raise CommandStringTooLongError(cmdLength)


class CommandStringTooLongError(Exception):

    def __init__(self, length):
        super(CommandStringTooLongError, self).__init__("Length " + str(length) + " too long (limit " + str(max_command_length) + ")")



import logging
import os
import subprocess
import Pyro4


def setHostname():
        ip = getHostname()
        logging.info("Using " + ip + " as hostname")
        Pyro4.config.HOST = ip


def getHostname():
    fromEnv = os.environ.get("PYRO_IP", False)
    if fromEnv:
        return fromEnv
    return subprocess.check_output(["hostname", "-I"]).split()[0].rstrip()



'''
Created on 10 Nov 2012

@author: james
'''
from avx.devices.serial import SerialDevice
from avx.Client import MessageTypes

import logging


class KramerVP88(SerialDevice):
    '''
    A Kramer VP88 switcher, controlled by Protocol 2000 over serial.
    '''

    def __init__(self, deviceID, serialDevice, machineNumber=1, **kwargs):
        super(KramerVP88, self).__init__(deviceID, serialDevice, **kwargs)
        self.machineNumber = machineNumber

    def sendInputToOutput(self, inChannel, outChannel):
        toSend = [0x01, 0x80 + int(inChannel), 0x80 + int(outChannel), 0x80 + self.machineNumber]
        return self.sendCommand(SerialDevice.byteArrayToString(toSend))

    def initialise(self):
        SerialDevice.initialise(self)
        self.port.flushInput()
        self.requestStatus()

    def requestStatus(self):
        for i in range(1, 9):
            self.sendCommand(SerialDevice.byteArrayToString([0x05, 0x80, 0x80 + i, 0x80 + self.machineNumber]))

    def hasFullMessage(self, recv_buffer):
        return len(recv_buffer) == 4

    def handleMessage(self, msgBytes):
        logging.debug("Received from Kramer VP-88: " + SerialDevice.byteArrayToString(msgBytes).encode('hex_codec'))
        print msgBytes, 0x80 + self.machineNumber
        if (msgBytes[3] == 0x80 + self.machineNumber):
            if (msgBytes[0] == 0x41) or (msgBytes[0] == 0x45):  # Notification of video switch or response to query
                inp = msgBytes[1] - 0x80
                outp = msgBytes[2] - 0x80
                self.broadcast(MessageTypes.OUTPUT_MAPPING, {outp: inp})



'''
Heavily borrowing from https://bitbucket.org/wump/ambx-python/ and https://code.google.com/p/combustd before it
'''

from avx.devices import Device
import usb
from array import array


# USB identification
VENDOR = 0x0471
PRODUCT = 0x083f

# Endpoints
EP_IN = 0x81
EP_OUT = 0x02
EP_PNP = 0x83

# Packet start
PKT_HEADER = 0xA1

# -- Commands --
SET_LIGHT_COLOR = 0x03
SET_TIMED_COLOR_SEQUENCE = 0x72  # Not used


# -- Lights --
class Lights:
    LEFT = 0x0B
    RIGHT = 0x1B
    WWLEFT = 0x2B
    WWCENTER = 0x3B
    WWRIGHT = 0x4B


class AMBX(Device):
    '''
    A Philips amBx USB device.
    '''

    def __init__(self, deviceID, deviceNum=0, **kwargs):
        super(AMBX, self).__init__(deviceID, **kwargs)
        # find our device
        devs = devices_by_vendor_product(VENDOR, PRODUCT)

        devptr = devs[deviceNum]

        self.dev = devptr.open()
        self.dev.claimInterface(0)

    def initialise(self):
        self.allOff()

    def setColour(self, light, red, green, blue):
        return self.dev.interruptWrite(EP_OUT, B([PKT_HEADER, light, SET_LIGHT_COLOR, red, green, blue]), 100)

    def allOff(self):
        self.setColour(Lights.LEFT, 0, 0, 0)
        self.setColour(Lights.WWLEFT, 0, 0, 0)
        self.setColour(Lights.WWCENTER, 0, 0, 0)
        self.setColour(Lights.WWRIGHT, 0, 0, 0)
        self.setColour(Lights.RIGHT, 0, 0, 0)


def devices_by_vendor_product(vendor, product):
    '''
    Enumerate all USB devices with the right vendor
    and product ID
    '''
    devs = []
    for bus in usb.busses():
        for device in bus.devices:
            if device.idVendor == vendor and device.idProduct == product:
                devs.append(device)
    return devs


def B(x):
    return array("B", x)



'''
Created on 13 Nov 2012

@author: jrem
'''
from avx.devices.serial import SerialDevice


class CoriogenEclipse(SerialDevice):
    '''
    Coriogen Eclipse scan converter/genlock/overlay
    '''

    def __init__(self, deviceID, serialDevice, **kwargs):
        super(CoriogenEclipse, self).__init__(deviceID, serialDevice, **kwargs)

    def overlayOn(self):
        self.sendCommand("Mode = 3\r\n")

    def overlayOff(self):
        self.sendCommand("Mode = 0\r\n")

    def fadeOut(self):
        self.sendCommand("Fade = 0\r\n")

    def fadeIn(self):
        self.sendCommand("Fade = 1\r\n")

    def freeze(self):
        self.sendCommand("Freeze = On\r\n")

    def unfreeze(self):
        self.sendCommand("Freeze = Off\r\n")

    def overscanOn(self):
        self.sendCommand("Overscan = On\r\n")

    def overscanOff(self):
        self.sendCommand("Overscan = Off\r\n")



from avx.devices.serial import SerialDevice


class KramerVP703(SerialDevice):
    '''
    Kramer VP703-SC scan converter.
    '''

    def __init__(self, deviceID, serialDevice, **kwargs):
        super(KramerVP703, self).__init__(deviceID, serialDevice, **kwargs)

    def initialise(self):
        super(KramerVP703, self).initialise()
        self.overscanOn()

    def overscanOn(self):
        return self.sendCommand("Overscan = 1\r\n")

    def overscanOff(self):
        return self.sendCommand("Overscan = 0\r\n")

    def freeze(self):
        return self.sendCommand("Image Freeze = 1\r\n")

    def unfreeze(self):
        return self.sendCommand("Image Freeze = 0\r\n")

    def recalibrate(self):
        return self.sendCommand("AutoTrack = 1\r\n")



from avx.devices import Device
import colorsys
import logging
import socket
import time


class MiLight(Device):
    '''
    Control of MiLight / easybulb wifi lightbulbs
    '''

    socket = None

    def __init__(self, deviceID, ipAddress, port=8899, **kwargs):
        super(MiLight, self).__init__(deviceID, **kwargs)
        self.ipAddress = ipAddress
        self.port = port

    def initialise(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def send(self, message):
        if not self.socket:
            self.initialise()
        logging.debug("Sending " + message.encode('hex_codec') + " to " + self.ipAddress + ":" + str(self.port))
        self.socket.sendto(message, (self.ipAddress, self.port))

    def sendGroupOnAndWait(self, group):
        self.send(chr(groupID(group)) + "\x00\x55")
        time.sleep(0.1)

    def setColour(self, group, red, green, blue):
        self.sendGroupOnAndWait(group)
        self.send("\x40" + chr(rgbToMilight(red, green, blue)) + "\x55")

    def setWhite(self, group):
        self.sendGroupOnAndWait(group)
        self.send(chr(groupID(group) + 0x80) + "\x00\x55")

    def setBrightness(self, group, percentage):
        '''
        Set the brightness of the group, in their current mode.
        Percentage should be [0-100].
        '''
        self.sendGroupOnAndWait(group)
        self.send("\x4E" + chr(2 + int(percentage / 4)) + "\x55")

    def setOff(self, group):
        if group == 0:
            self.send("\x41\x00\x55")
        else:
            self.send(chr(groupID(group) + 1) + "\x00\x55")


def groupID(group):
    if group is 0:
        return 0x42
    else:
        return 67 + (2 * group)


def rgbToMilight(red, green, blue):
    '''
    This isn't quite right, and you can't do a proper conversion since milight requires a saturation of 1.
    Also, we throw away the value - we just extract the hue of the rgb colour and try to map that to the
    milight colour chart.
    '''
    hsv = colorsys.rgb_to_hsv(red, green, blue)
    hue = hsv[0] * 360
    milight = ((250 - hue) % 360) * 256 / 360
    return int(round(milight))



from argparse import ArgumentParser, FileType
from avx import PyroUtils, _version
from avx.controller.ControllerHttp import ControllerHttp
from avx.devices import Device
from avx.Sequencer import Sequencer
from logging import Handler
from Pyro4.errors import NamingError
from semantic_version import Version as SemVer
import atexit
import logging.config
import Pyro4
import json

Pyro4.config.SERIALIZER = 'pickle'
Pyro4.config.SERIALIZERS_ACCEPTED.add('pickle')
Pyro4.config.REQUIRE_EXPOSE = False


def versionsCompatible(remote, local):
    rv = SemVer(remote, partial=True)
    lv = SemVer(local, partial=True)
    if rv.major == 0:
        return rv.major == lv.major and rv.minor == lv.minor
    return rv.major == lv.major and rv.minor >= lv.minor


class Controller(object):
    '''
    A Controller is essentially a bucket of devices, each identified with a string deviceID.
    '''
    pyroName = "avx.controller"
    version = _version.__version__

    def __init__(self):
        self.devices = {}
        self.proxies = {}
        self.sequencer = Sequencer(self)
        self.sequencer.start()
        self.logHandler = ControllerLogHandler()
        logging.getLogger().addHandler(self.logHandler)
        self.clients = []
        self.slaves = []
        self.daemon = Pyro4.Daemon(PyroUtils.getHostname())

    @staticmethod
    def fromPyro(controllerID=""):
        controllerAddress = "PYRONAME:" + Controller.pyroName
        if controllerID != "":
            controllerAddress += "." + controllerID
        logging.info("Creating proxy to controller at " + controllerAddress)

        controller = ControllerProxy(Pyro4.Proxy(controllerAddress))
        remoteVersion = controller.getVersion()
        if not versionsCompatible(remoteVersion, Controller.version):
            raise VersionMismatchError(remoteVersion, Controller.version)
        return controller

    def loadConfig(self, configFile, overrideToDebug=False):
        try:
            if isinstance(configFile, file):
                config = json.load(configFile)
                self.configFile = configFile.name
            else:
                config = json.load(open(configFile))
                self.configFile = configFile

            self.config = config

            for d in config["devices"]:
                device = Device.create(d, self)
                self.addDevice(device)

            if "options" in config:

                if "controllerID" in config["options"]:
                    self.controllerID = config["options"]["controllerID"]

                if "slaves" in config["options"]:
                    for slave in config["options"]["slaves"]:
                        try:
                            sc = Controller.fromPyro(slave)

                            if versionsCompatible(sc.getVersion(), self.getVersion()):
                                self.slaves.append(sc)
                            else:
                                logging.error("This Controller is version " + str(self.getVersion()) + " but tried to add slave " + slave + " of version " + str(sc.getVersion()))
                        except NamingError:
                            logging.error("Could not connect to slave with controller ID " + slave)

                if "http" in config["options"]:
                    if config["options"]["http"] is True:
                        ch = ControllerHttp(self)
                        ch.start()

            if "logging" in config:
                logging.config.dictConfig(config["logging"])
                if overrideToDebug:
                    logging.getLogger().setLevel(logging.DEBUG)
                    logging.info("-d specified, overriding any specified default logger level to DEBUG")

            if "clients" in config:
                self.clients = list(config["clients"])
            self.config["clients"] = self.clients

        except ValueError:
            logging.exception("Cannot parse config.json!")

    def saveConfig(self):
        if hasattr(self, "configFile"):
            try:
                with open(self.configFile, "w") as cfout:
                    json.dump(self.config, cfout)
            except IOError:
                logging.exception("Cannot save config file!")

    def registerClient(self, clientURI):
        self.clients.append(str(clientURI))
        logging.info("Registered client at " + str(clientURI))
        logging.info(str(len(self.clients)) + " client(s) now connected")
        self.saveConfig()

    def unregisterClient(self, clientURI):
        self.clients.remove(str(clientURI))
        logging.info("Unregistered client at " + str(clientURI))
        logging.info(str(len(self.clients)) + " client(s) still connected")
        self.saveConfig()

    def broadcast(self, msgType, source, data):
        ''' Send a message to all clients '''
        logging.info("Broadcast: {}, {}, {}".format(msgType, source, data))
        for uri in list(self.clients):
            try:
                logging.debug("Calling handleMessage on client at {}".format(uri))
                client = Pyro4.Proxy(uri)
                result = client.handleMessage(msgType, source, data)
                logging.debug("Client call returned " + str(result))
            except Exception:
                logging.exception("Failed to call handleMessage on registered client {}, removing.".format(uri))
                self.unregisterClient(uri)

    def getVersion(self):
        return self.version

    def addDevice(self, device):
        if self.hasDevice(device.deviceID):
            raise DuplicateDeviceIDError(device.deviceID)
        self.devices[device.deviceID] = device
        if hasattr(device, "registerDispatcher") and callable(getattr(device, "registerDispatcher")):
            device.registerDispatcher(self)
        device.broadcast = lambda t, b: self.broadcast(t, device.deviceID, b)

    def getDevice(self, deviceID):
        return self.devices[deviceID]

    def proxyDevice(self, deviceID):
        if deviceID not in self.proxies.keys():
            if self.hasDevice(deviceID):
                self.proxies[deviceID] = self.daemon.register(self.getDevice(deviceID))
            else:
                for slave in self.slaves:
                    if slave.hasDevice(deviceID):
                        self.proxies[deviceID] = slave.proxyDevice(deviceID)
        return self.proxies[deviceID]

    def hasDevice(self, deviceID):
        return deviceID in self.devices

    def initialise(self):
        for device in self.devices.itervalues():
            device.initialise()
        atexit.register(self.deinitialise)

    def deinitialise(self):
        for device in self.devices.itervalues():
            device.deinitialise()

    def startServing(self):
        PyroUtils.setHostname()

        ns = Pyro4.locateNS()
        uri = self.daemon.register(self)

        if hasattr(self, "controllerID"):
            name = self.pyroName + "." + self.controllerID
        else:
            name = self.pyroName

        logging.info("Registering controller as " + name)

        ns.register(name, uri)

        atexit.register(lambda: self.daemon.shutdown())

        self.daemon.requestLoop()

    def sequence(self, *events):
        self.sequencer.sequence(*events)

    def getLog(self):
        return self.logHandler.entries


class ControllerProxy(object):
    def __init__(self, controller):
        self.controller = controller

    def __getattr__(self, name):
        return getattr(self.controller, name)

    def __getitem__(self, item):
        return Pyro4.Proxy(self.controller.proxyDevice(item))


class ControllerLogHandler(Handler):

    def __init__(self):
        Handler.__init__(self)
        self.entries = []

    def emit(self, record):
        self.entries.append(record)
        if len(self.entries) > 100:
            self.entries.pop(0)
        if record.exc_info is not None:
            record.exc_info = None
            fakeRecord = logging.LogRecord("Controller", logging.WARNING, record.pathname, record.lineno, "", {}, None, None)
            fakeRecord.created = record.created
            fakeRecord.asctime = record.asctime if hasattr(record, "asctime") else "--"
            self.format(fakeRecord)
            fakeRecord.message = "An exception was stripped from this log, see controller logs for details"
            self.entries.append(fakeRecord)


class VersionMismatchError(Exception):

    def __init__(self, remoteVersion, localVersion):
        super(VersionMismatchError, self).__init__("Controller is version " + str(remoteVersion) + " but this client is written for version " + str(localVersion) + ". Check your installation and try again.")


class DuplicateDeviceIDError(Exception):

    def __init__(self, duplicatedID):
        super(DuplicateDeviceIDError, self).__init__("Device already exists: " + duplicatedID)


def main():
    parser = ArgumentParser()
    parser.add_argument("-d", "--debug",
                        help="Show debugging output.",
                        action="store_true")
    parser.add_argument("-c", "--config",
                        help="Configuration file to use",
                        type=FileType("r"))
    args = parser.parse_args()
    logging.basicConfig(format='%(asctime)s %(levelname)s: %(message)s', level=(logging.DEBUG if args.debug else logging.INFO))

    controller = Controller()
    logging.info("Starting avx controller v{}".format(controller.getVersion()))

    if args.config:
        controller.loadConfig(args.config, args.debug)
    else:
        try:
            configFile = open('config.json', 'r')
            controller.loadConfig(configFile)
        except IOError:
            logging.error("No config file specified and config.json not found! Exiting...")
            exit(1)
    controller.initialise()
    controller.startServing()
    logging.info("avx controller terminated.")


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages
import re

VERSIONFILE = "src/avx/_version.py"
verstr = "unknown"
try:
    verstrline = open(VERSIONFILE, "rt").read()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(VSRE, verstrline, re.M)
    if mo:
        verstr = mo.group(1)
except EnvironmentError:
    print "unable to find version in %s" % (VERSIONFILE,)
    raise RuntimeError("if %s exists, it is required to be well-formed" % (VERSIONFILE,))

setup(
    name='avx',
    version=verstr,
    description='Library for controlling A/V devices',
    author='James Muscat',
    author_email='jamesremuscat@gmail.com',
    url='https://github.com/jamesremuscat/avx',
    packages=find_packages('src', exclude=["*.tests"]),
    package_dir = {'':'src'},
      long_description="""\
      AVX is a library for controlling A/V devices such as video switchers. ...
      """,
    setup_requires=["nose>=1.0"],
    tests_require=["mock"],
    install_requires=["enum34", "Pyro4>=4.20,!=4.45", "pyserial", "pyusb", "semantic_version"],
    entry_points={
        'console_scripts': [
            'avx-controller = avx.controller.Controller:main',
            ],
        }
      )



'''
Created on 13 Nov 2012

@author: james
'''
from avx.devices import InvalidArgumentException
from avx.devices.serial import SerialDevice
from avx.CameraPosition import CameraPosition
from enum import Enum
from threading import Lock, ThreadError, Event

import logging


# Pan speeds can vary from 0x01 - 0x18
DEFAULT_PAN_SPEED = 0x06
# Tilt speeds can vary from 0x01 - 0x14
DEFAULT_TILT_SPEED = 0x06
# Zoom speed can vary from 0x02-0x07
DEFAULT_ZOOM_SPEED = 0x06


def constrainPanTiltSpeed(func):
    def inner(elf, panSpeed=DEFAULT_PAN_SPEED, tiltSpeed=DEFAULT_TILT_SPEED):
        checkPan(panSpeed)
        checkTilt(tiltSpeed)
        func(elf, panSpeed, tiltSpeed)
    return inner


class VISCACamera(SerialDevice):
    '''
    A camera controlled by the Sony VISCA protocol e.g. Sony D31.

    Limitation: 'proper' VISCA requires waiting for an ACK/NACK from camera
    before sending next command. We completely ignore ACKs and NACKs and just
    spew commands as often as we're asked to.
    '''

    def __init__(self, deviceID, serialDevice, cameraID, **kwargs):
        super(VISCACamera, self).__init__(deviceID, serialDevice, **kwargs)
        self.cameraID = cameraID
        self._wait_for_ack = Lock()
        self._wait_for_response = Lock()
        self._response_received = Event()
        self._last_response = None

    def sendVISCA(self, commandBytes):
        self._wait_for_ack.acquire()
        return self.sendCommand(SerialDevice.byteArrayToString([0x80 + self.cameraID] + commandBytes + [0xFF]))

    def getVISCA(self, commandBytes):
        with self._wait_for_response:
            self.sendVISCA(commandBytes)
            logging.debug("Waiting for response.")
            self._response_received.wait()
            logging.debug("Received response")
            response = self._last_response
            self._response_received.clear()
            return response

    @constrainPanTiltSpeed
    def moveUp(self, pan=DEFAULT_PAN_SPEED, tilt=DEFAULT_TILT_SPEED):
        return self.sendVISCA([0x01, 0x06, 0x01, pan, tilt, 0x03, 0x01])

    @constrainPanTiltSpeed
    def moveUpLeft(self, pan=DEFAULT_PAN_SPEED, tilt=DEFAULT_TILT_SPEED):
        return self.sendVISCA([0x01, 0x06, 0x01, pan, tilt, 0x01, 0x01])

    @constrainPanTiltSpeed
    def moveLeft(self, pan=DEFAULT_PAN_SPEED, tilt=DEFAULT_TILT_SPEED):
        return self.sendVISCA([0x01, 0x06, 0x01, pan, tilt, 0x01, 0x03])

    @constrainPanTiltSpeed
    def moveDownLeft(self, pan=DEFAULT_PAN_SPEED, tilt=DEFAULT_TILT_SPEED):
        return self.sendVISCA([0x01, 0x06, 0x01, pan, tilt, 0x01, 0x02])

    @constrainPanTiltSpeed
    def moveDown(self, pan=DEFAULT_PAN_SPEED, tilt=DEFAULT_TILT_SPEED):
        return self.sendVISCA([0x01, 0x06, 0x01, pan, tilt, 0x03, 0x02])

    @constrainPanTiltSpeed
    def moveDownRight(self, pan=DEFAULT_PAN_SPEED, tilt=DEFAULT_TILT_SPEED):
        return self.sendVISCA([0x01, 0x06, 0x01, pan, tilt, 0x02, 0x02])

    @constrainPanTiltSpeed
    def moveRight(self, pan=DEFAULT_PAN_SPEED, tilt=DEFAULT_TILT_SPEED):
        return self.sendVISCA([0x01, 0x06, 0x01, pan, tilt, 0x02, 0x03])

    @constrainPanTiltSpeed
    def moveUpRight(self, pan=DEFAULT_PAN_SPEED, tilt=DEFAULT_TILT_SPEED):
        return self.sendVISCA([0x01, 0x06, 0x01, pan, tilt, 0x02, 0x01])

    @constrainPanTiltSpeed
    def stop(self, pan=DEFAULT_PAN_SPEED, tilt=DEFAULT_TILT_SPEED):
        return self.sendVISCA([0x01, 0x06, 0x01, pan, tilt, 0x03, 0x03])

    def zoomIn(self, speed=DEFAULT_ZOOM_SPEED):
        checkZoom(speed)
        return self.sendVISCA([0x01, 0x04, 0x07, 0x20 + speed])

    def zoomOut(self, speed=DEFAULT_ZOOM_SPEED):
        checkZoom(speed)
        return self.sendVISCA([0x01, 0x04, 0x07, 0x30 + speed])

    def zoomStop(self):
        return self.sendVISCA([0x01, 0x04, 0x07, 0x00])

    def focusFar(self):
        self.focusManual()
        return self.sendVISCA([0x01, 0x04, 0x08, 0x02])

    def focusNear(self):
        self.focusManual()
        return self.sendVISCA([0x01, 0x04, 0x08, 0x03])

    def focusStop(self):
        return self.sendVISCA([0x01, 0x04, 0x08, 0x00])

    def focusAuto(self):
        return self.sendVISCA([0x01, 0x04, 0x38, 0x02])

    def focusManual(self):
        return self.sendVISCA([0x01, 0x04, 0x38, 0x03])

    def brighter(self):
        self.sendVISCA([0x01, 0x04, 0x39, 0x0D])  # Put camera into manual exposure mode first!
        return self.sendVISCA([0x01, 0x04, 0x0D, 0x02])

    def darker(self):
        self.sendVISCA([0x01, 0x04, 0x39, 0x0D])  # Put camera into manual exposure mode first!
        return self.sendVISCA([0x01, 0x04, 0x0D, 0x03])

    def autoExposure(self):
        return self.sendVISCA([0x01, 0x04, 0x39, 0x00])

    def backlightCompOn(self):
        return self.sendVISCA([0x01, 0x04, 0x33, 0x02])

    def backlightCompOff(self):
        return self.sendVISCA([0x01, 0x04, 0x33, 0x03])

    def storePreset(self, preset):
        if preset < 0 or preset > 5:
            return -1
        self.sendVISCA([0x01, 0x04, 0x3F, 0x01, preset])

    def recallPreset(self, preset):
        if preset < 0 or preset > 5:
            return -1
        self.sendVISCA([0x01, 0x04, 0x3F, 0x02, preset])

    def whiteBalanceAuto(self):
        self.sendVISCA([0x01, 0x04, 0x35, 0x00])

    def whiteBalanceIndoor(self):
        self.sendVISCA([0x01, 0x04, 0x35, 0x01])

    def whiteBalanceOutdoor(self):
        self.sendVISCA([0x01, 0x04, 0x35, 0x02])

    def whiteBalanceOnePush(self):
        self.sendVISCA([0x01, 0x04, 0x35, 0x03])

    def whiteBalanceOnePushTrigger(self):
        self.sendVISCA([0x01, 0x04, 0x10, 0x05])

    def setAutoExposure(self):
        self.sendVISCA([0x01, 0x04, 0x39, 0x00])

    def setAperturePriority(self):
        self.sendVISCA([0x01, 0x04, 0x39, 0x0B])

    def setShutterPriority(self):
        self.sendVISCA([0x01, 0x04, 0x39, 0x0A])

    def setManualExposure(self):
        self.sendVISCA([0x01, 0x04, 0x39, 0x03])

    def setAperture(self, aperture):
        if isinstance(aperture, Aperture):
            av = aperture.code
        else:
            av = aperture
        self.sendVISCA([0x01, 0x04, 0x4B,
                        (av & 0xF000) >> 12,
                        (av & 0x0F00) >> 8,
                        (av & 0x00F0) >> 4,
                        (av & 0x000F)])

    def setShutter(self, shutter):
        if isinstance(shutter, Shutter):
            tv = shutter.code
        else:
            tv = shutter
        self.sendVISCA([0x01, 0x04, 0x4A,
                        (tv & 0xF000) >> 12,
                        (tv & 0x0F00) >> 8,
                        (tv & 0x00F0) >> 4,
                        (tv & 0x000F)])

    def setGain(self, gain):
        if isinstance(gain, Gain):
            h = gain.code
        else:
            h = gain
        self.sendVISCA([0x01, 0x04, 0x4C,
                        (h & 0xF000) >> 12,
                        (h & 0x0F00) >> 8,
                        (h & 0x00F0) >> 4,
                        (h & 0x000F)])

    def hasFullMessage(self, recv_buffer):
        return len(recv_buffer) > 0 and recv_buffer[-1] == 0xFF

    def handleMessage(self, msgBytes):
        responseType = (msgBytes[1] & 0x70) >> 4
        logging.debug("Response of type {}".format(responseType))
        if responseType >= 4 and responseType <= 6:  # ack, response or error
            try:
                self._wait_for_ack.release()
            except ThreadError:
                # We weren't blocked anyway
                pass
        if responseType == 5:
            self._last_response = msgBytes
            self._response_received.set()

    def getPosition(self):
        # returns Y0 50 0W 0W 0W 0W 0Z 0Z 0Z 0Z FF where WWWW = pan, ZZZZ = tilt
        pan_tilt_resp = self.getVISCA([0x09, 0x06, 0x12, 0xFF])

        # returns Y0 50 0Z 0Z 0Z 0Z FF where ZZZZ = zoom
        zoom_resp = self.getVISCA([0x09, 0x04, 0x47, 0xFF])

        pan = (pan_tilt_resp[2] << 12) + (pan_tilt_resp[3] << 8) + (pan_tilt_resp[4] << 4) + pan_tilt_resp[5]
        tilt = (pan_tilt_resp[6] << 12) + (pan_tilt_resp[7] << 8) + (pan_tilt_resp[8] << 4) + pan_tilt_resp[9]

        zoom = (zoom_resp[2] << 12) + (zoom_resp[3] << 8) + (zoom_resp[4] << 4) + zoom_resp[5]

        return CameraPosition(pan, tilt, zoom)

    def goto(self, pos, DEFAULT_PAN_SPEED, DEFAULT_TILT_SPEED):
        '''
        Takes a CameraPosition to directly set the required tilt, pan and zoom of the camera, and the speed at which to get there.
        VISCA somewhat arbitrarily limits these values to:
         - pan range: 0xFC90 to 0x0370 (centred on 0)
         - pan speed: 0x01-0x18
         - tilt range: 0xFED4 to 0x012C (centred on 0)
         - tilt speed: 0x01 - 0x14
         - zoom value: 0x00 (wide) to 0x03FF (tele)
        '''
        p = pos.pan
        t = pos.tilt
        z = pos.zoom

        setPT = [
            0x01,
            0x06,
            0x02,
            DEFAULT_PAN_SPEED & 0xFF,
            DEFAULT_TILT_SPEED & 0xFF,
            # Pan x 2 bytes, padded to four (ABCD -> 0A 0B 0C 0D)
            (p & 0xF000) >> 12,
            (p & 0x0F00) >> 8,
            (p & 0x00F0) >> 4,
            (p & 0x000F),
            # Tilt x 2 bytes, padded to four (ABCD -> 0A 0B 0C 0D)
            (t & 0xF000) >> 12,
            (t & 0x0F00) >> 8,
            (t & 0x00F0) >> 4,
            (t & 0x000F),
            0xFF,
        ]

        ret = self.sendVISCA(setPT)

        setZ = [
            0x01,
            0x04,
            0x47,
            (z & 0xF000) >> 12,
            (z & 0x0F00) >> 8,
            (z & 0x00F0) >> 4,
            (z & 0x000F),
            0xFF,
        ]

        ret += self.sendVISCA(setZ)

        return ret


class Aperture(Enum):
    CLOSE = (0x00, "Closed")
    F28 = (0x01, "F28")
    F22 = (0x02, "F22")
    F19 = (0x03, "F19")
    F16 = (0x04, "F16")
    F14 = (0x05, "F14")
    F11 = (0x06, "F11")
    F9_6 = (0x07, "F9.6")
    F8 = (0x08, "F8")
    F6_8 = (0x09, "F6.8")
    F5_6 = (0x0A, "F5.6")
    F4_8 = (0x0B, "F4.8")
    F4 = (0x0C, "F4")
    F3_4 = (0x0D, "F3.4")
    F2_8 = (0x0E, "F2.8")
    F2_4 = (0x0F, "F2.4")
    F2 = (0x10, "F2")
    F1_8 = (0x11, "F1.8")

    def __init__(self, code, label):
        self.code = code
        self.label = label


class Shutter(Enum):
    T50 = (0x00, "1/50s")
    T60 = (0x01, "1/60s")
    T75 = (0x02, "1/75s")
    T90 = (0x03, "1/90s")
    T100 = (0x04, "1/100s")
    T120 = (0x05, "1/120s")
    T150 = (0x06, "1/150s")
    T180 = (0x07, "1/180s")
    T215 = (0x08, "1/215s")
    T250 = (0x09, "1/250s")
    T300 = (0x0A, "1/300s")
    T350 = (0x0B, "1/350s")
    T425 = (0x0C, "1/425s")
    T500 = (0x0D, "1/500s")
    T600 = (0x0E, "1/600s")
    T725 = (0x0F, "1/725s")
    T850 = (0x10, "1/850s")
    T1000 = (0x11, "1/1000s")
    T1250 = (0x12, "1/1250s")
    T1500 = (0x13, "1/1500s")
    T1750 = (0x14, "1/1750s")
    T2000 = (0x15, "1/2000s")
    T2500 = (0x16, "1/2500s")
    T3000 = (0x17, "1/3000s")
    T3500 = (0x18, "1/3500s")
    T4000 = (0x19, "1/4000s")
    T6000 = (0x1A, "1/6000s")
    T10000 = (0x1B, "1/10000s")

    def __init__(self, code, label):
        self.code = code
        self.label = label


class Gain(Enum):
    G_MINUS_3 = (0x00, "-3")
    G_0 = (0x01, "0")
    G_3 = (0x02, "3")
    G_6 = (0x03, "6")
    G_9 = (0x04, "9")
    G_12 = (0x05, "12")
    G_15 = (0x06, "15")
    G_18 = (0x07, "18")

    def __init__(self, code, label):
        self.code = code
        self.label = label


def checkPan(pan):
    if pan < 1 or pan > 0x18:
        raise InvalidArgumentException()


def checkTilt(tilt):
    if tilt < 1 or tilt > 0x16:
        raise InvalidArgumentException()


def checkZoom(zoom):
    if zoom < 2 or zoom > 7:
        raise InvalidArgumentException()



'''
Created on 8 Apr 2013

@author: jrem
'''
from avx import PyroUtils

import Pyro4
import atexit
import threading


class Client(threading.Thread):

    def __init__(self):
        super(Client, self).__init__()
        self.started = threading.Event()

    def run(self):
        PyroUtils.setHostname()
        daemon = Pyro4.Daemon()
        self.uri = daemon.register(self)
        self.started.set()
        atexit.register(lambda: daemon.shutdown())
        daemon.requestLoop()

    def getHostIP(self):
        return Pyro4.config.HOST

    @Pyro4.expose
    def handleMessage(self, msgType, sourceDeviceID, data):
        pass


class MessageTypes(object):
    OUTPUT_MAPPING = "avx.client.OutputMapping"



from avx.devices.serial import SerialDevice
from avx.devices import Device, InvalidArgumentException
from time import sleep
from threading import Thread, Lock

import logging
import time


class SerialRelayCard(SerialDevice):
    '''
    A serial relay card is a single serial device which may control multiple relays.
    For convenience, each relay on a card may have its own device which can be created by calling
    createDevice on the relay card.
    '''

    def __init__(self, deviceID, serialDevice, **others):
        SerialDevice.__init__(self, deviceID, serialDevice, **others)

    def createDevice(self, deviceID, channel):
        return RelayDevice(deviceID, self, channel)


class RelayDevice(Device):

    def __init__(self, deviceID, relayCard, channel, **kwargs):
        super(RelayDevice, self).__init__(deviceID, **kwargs)
        self.relayCard = relayCard
        self.channel = channel

    def on(self):
        return self.relayCard.on(self.channel)

    def off(self):
        return self.relayCard.off(self.channel)


class KMtronicSerialRelayCard(SerialRelayCard):

    def __init__(self, deviceID, serialDevice, **kwargs):
        SerialRelayCard.__init__(self, deviceID, serialDevice, **kwargs)

    def on(self, channel):
        return self.sendCommand(SerialDevice.byteArrayToString([0xFF, channel, 0x01]))

    def off(self, channel):
        return self.sendCommand(SerialDevice.byteArrayToString([0xFF, channel, 0x00]))


class JBSerialRelayCard(SerialRelayCard):

    sendDelay = 0.05

    def __init__(self, deviceID, serialDevice, **kwargs):
        SerialRelayCard.__init__(self, deviceID, serialDevice, baud=19200, **kwargs)

    def on(self, channel):
        result = self.sendCommand(SerialDevice.byteArrayToString([0x30 + 2 * channel]))
        time.sleep(self.sendDelay)
        return result

    def off(self, channel):
        result = self.sendCommand(SerialDevice.byteArrayToString([0x31 + 2 * channel]))
        time.sleep(self.sendDelay)
        return result


class ICStationSerialRelayCard(SerialRelayCard):

    def __init__(self, deviceID, serialDevice, channels=8, **kwargs):
        SerialRelayCard.__init__(self, deviceID, serialDevice, **kwargs)
        self.state = [False for _ in range(channels)]  # True = on, False = off
        self.initialised = False
        self.run_send_thread = False
        self.prevState = None

    def initialise(self):
        if not self.initialised:
            SerialRelayCard.initialise(self)
            self.sendCommand("\x50")
            time.sleep(0.1)
            self.sendCommand("\x51")
            self.initialised = True
        self.run_send_thread = True

        def run():
            while self.run_send_thread:
                self._sendStateCommand()
                sleep(0.1)

        runner = Thread(target=run)
        runner.setDaemon(True)
        runner.start()

    def deinitialise(self):
        self.run_send_thread = False

    def _sendStateCommand(self):
        toSend = self._createStateByte()
        if toSend != self.prevState:
            self.prevState = toSend
            result = self.sendCommand(SerialDevice.byteArrayToString([toSend]))
            return result

    def _createStateByte(self):
        stateByte = 0x0
        for i in range(0, len(self.state)):
            if not self.state[i]:  # Card requires bit = 0 to turn relay on
                stateByte += 1 << i
        return stateByte

    def _checkSenderState(self):
        if not self.run_send_thread:
            logging.warn("State change requested for {} but run_send_thread is False - device uninitialised?".format(self.deviceID))

    def on(self, channel):
        self.__checkChannel(channel)
        self._checkSenderState()
        self.state[channel - 1] = True

    def off(self, channel):
        self.__checkChannel(channel)
        self._checkSenderState()
        self.state[channel - 1] = False

    def __checkChannel(self, channel):
        if channel > len(self.state):
            raise InvalidArgumentException()


class UpDownStopRelay(Device):

    def __init__(self, deviceID, controller, directionRelay, startStopRelay, **kwargs):
        super(UpDownStopRelay, self).__init__(deviceID, **kwargs)
        self.directionRelay = controller.getDevice(directionRelay[0]).createDevice(self.deviceID + "_direction", directionRelay[1])
        self.startStopRelay = controller.getDevice(startStopRelay[0]).createDevice(self.deviceID + "_startStop", startStopRelay[1])

    def raiseUp(self):
        self.directionRelay.on()
        self.startStopRelay.on()

    def lower(self):
        self.directionRelay.off()
        self.startStopRelay.on()

    def stop(self):
        self.startStopRelay.off()


class MomentaryUpDownStopRelay(Device):
    def __init__(self, deviceID, controller, upRelay, downRelay, stopRelay, **kwargs):
        super(MomentaryUpDownStopRelay, self).__init__(deviceID, **kwargs)
        self.upRelay = controller.getDevice(upRelay[0]).createDevice(self.deviceID + "_up", upRelay[1])
        self.downRelay = controller.getDevice(downRelay[0]).createDevice(self.deviceID + "_down", downRelay[1])
        self.stopRelay = controller.getDevice(stopRelay[0]).createDevice(self.deviceID + "_stop", stopRelay[1])
        self.lock = Lock()

    def _touch(self, relay):
        # We don't want to block the calling thread, so waiting for our lock happens in its own thread.
        # Thus, UpDownStopArray still works quickly, but you can't force multiple relays to be on for a single MUDS.
        def inner():
            with self.lock:
                relay.on()
                time.sleep(0.5)
                relay.off()
        Thread(target=inner).start()

    def raiseUp(self):
        self._touch(self.upRelay)

    def lower(self):
        self._touch(self.downRelay)

    def stop(self):
        self._touch(self.stopRelay)


class UpDownStopArray(Device):

    def __init__(self, deviceID, controller, relays={}, **kwargs):
        super(UpDownStopArray, self).__init__(deviceID, **kwargs)
        self.relays = {}
        for idx, devID in relays.iteritems():
            self.relays[int(idx)] = controller.getDevice(devID)

    def add(self, device, number):
        self.relays[number] = device

    def raiseUp(self, number):
        if number in self.relays.keys():
            self.relays[number].raiseUp()
        elif number == 0:
            for r in self.relays.values():
                r.raiseUp()
        else:
            logging.error("Tried to raise relay channel " + str(number) + " but no such device attached to " + self.deviceID)

    def lower(self, number):
        if number in self.relays.keys():
            self.relays[number].lower()
        elif number == 0:
            for r in self.relays.values():
                r.lower()
        else:
            logging.error("Tried to lower relay channel " + str(number) + " but no such device attached to " + self.deviceID)

    def stop(self, number):
        if number in self.relays.keys():
            self.relays[number].stop()
        elif number == 0:
            for r in self.relays.values():
                r.stop()
        else:
            logging.error("Tried to stop relay channel " + str(number) + " but no such device attached to " + self.deviceID)



from avx.devices import Device
from serial import Serial, SerialException
from threading import Thread

import logging
import time


class SerialDevice(Device):
    '''
    A device we connect to over a serial port.
    '''

    def __init__(self, deviceID, serialDevice, baud=9600, **kwargs):
        '''
        Create a SerialDevice with a given device ID, using a new Serial with the given device address
        (e.g. "/dev/ttyUSB0") if given as a string, else assume serialDevice is a Serial and use that.
        If we fail to create a Serial with the given device, log an error and use a fake port.
        '''
        super(SerialDevice, self).__init__(deviceID, **kwargs)
        if isinstance(serialDevice, str) or isinstance(serialDevice, unicode):
            try:
                self.port = Serial(serialDevice, baud, timeout=2)
            except (SerialException, OSError):
                logging.exception("Could not open serial device " + serialDevice + " for " + deviceID)
                self.port = FakeSerialPort()
        else:
            self.port = serialDevice
        self.recv_thread = None
        self.dispatchers = []

    def initialise(self):
        self.recv_buffer = []
        self.run_receive = True
        if not (self.recv_thread and self.recv_thread.is_alive()):
            self.recv_thread = Thread(target=self._receive)
            self.recv_thread.daemon = True
            self.recv_thread.start()

    def _receive(self):
        while self.run_receive:
            read_byte = self.port.read(1)
            if read_byte:
                self.recv_buffer.append(ord(read_byte))
                if self.hasFullMessage(self.recv_buffer[:]):
                    logging.debug("Received from {}: {}".format(self.deviceID, SerialDevice.byteArrayToString(self.recv_buffer).encode('hex_codec')))
                    self.handleMessage(self.recv_buffer[:])
                    self.recv_buffer = []

    def hasFullMessage(self, recv_buffer):
        # Default implementation: always return True. This is almost certainly not useful to you so override this!
        # This a a safe default which, combined with handleMessage below, means that if you don't care about
        # receiving messages then you can just forget about overriding the two methods.
        return True

    def handleMessage(self, msgBytes):
        # Default implementation: just ignore messages.
        pass

    def registerDispatcher(self, dispatcher):
        self.dispatchers.append(dispatcher)

    def sendCommand(self, commandString):
        try:
            logging.debug("Sending " + commandString.encode('hex_codec') + " to " + self.port.portstr)
            sentBytes = self.port.write(commandString)
            logging.debug(str(sentBytes) + " bytes sent")
            return sentBytes
        except SerialException:
            #  Retry just the once - experience suggests that this might help
            logging.warn("Failed sending command to " + self.deviceID + " on " + self.port.portstr + ", retrying...")
            try:
                self.port.close()
                self.port.open()
                sentBytes = self.port.write(commandString)
                logging.debug(str(sentBytes) + " bytes sent")
                return sentBytes
            except SerialException:
                logging.exception("Really failed sending command to " + self.deviceID + " on " + self.port.portstr)

    def deinitialise(self):
        self.run_receive = False
        self.port.close()

    @staticmethod
    def byteArrayToString(byteArray):
        return ''.join(chr(b) for b in byteArray)


class FakeSerialPort(object):
    '''
    A class that quacks like a Serial if you try and write to it, but just throws everything away.
    '''
    portstr = "FAKE"

    def write(self, stuff):
        return 0

    def read(self, length):
        time.sleep(1)
        return []

    def flushInput(self):
        return 0

    def close(self):
        pass



'''
Created on 13 Nov 2012

@author: jrem
'''
from avx.devices.serial import SerialDevice
from avx.Client import MessageTypes

import logging


class Kramer602(SerialDevice):
    '''
    The Kramer 602 (preview) switcher. Reverse-engineered from the AMX code, this is apparently not Protocol 2000...
    '''

    def __init__(self, deviceID, serialDevice, machineNumber=1, **kwargs):
        super(Kramer602, self).__init__(deviceID, serialDevice, 1200, **kwargs)
        self.machineNumber = machineNumber

    def sendInputToOutput(self, inChannel, outChannel):
        if outChannel > 2:
            logging.error("Output channel " + str(outChannel) + " does not exist on switcher " + self.deviceID)
        else:
            code = [0, 0x80 + (2 * inChannel) - (2 - outChannel)]
            self.sendCommand(SerialDevice.byteArrayToString(code))

    def initialise(self):
        SerialDevice.initialise(self)
        self.port.flushInput()
        self.requestStatus()

    '''Request the status of this switcher's outputs.'''
    def requestStatus(self):
        self.port.write("\x00\xA1")

    def hasFullMessage(self, recv_buffer):
        return (len(recv_buffer) == 2 and recv_buffer[1] != 0xFF) or len(recv_buffer) == 3

    def handleMessage(self, msgBytes):
        if len(msgBytes) == 3:
            msgBytes.pop(1)  # Get rid of the mysterious middle byte
        if ((msgBytes[0] & 0x7) == (self.machineNumber - 1)):
            if (msgBytes[1] & 0x20) == 0:  # Not just a "I switched successfully" message
                outp = (((msgBytes[1] & 0x1F) - 1) % 2) + 1
                inp = (((msgBytes[1] & 0x1F) - outp) / 2) + 1  # int(math.ceil((message[1] & 0x1F) + (2 / 2)) - 1)
                self.broadcast(MessageTypes.OUTPUT_MAPPING, {outp: inp})
            else:
                self.requestStatus()  # Request device to send current status so listener can intercept
